# -*- coding: utf8 -*-
from __future__ import unicode_literals

import time
import logging
from driverng import Worklist
from driverng.astm import (
    AstmMessageV2,
    Segment,
    NoRangSegment,
    HeaderSegment,
    HprimString as HL7String,
    HprimDate as HL7Date,
    HprimForeign as HL7Foreign,
)

logger = logging.getLogger("driverng.atellica800")


class Au800CytologieSegmentMSH(HeaderSegment):
    sending_application = HL7String(3)
    date = HL7Date(7)
    message_type = HL7String((9, 1), default="OML")
    message_code = HL7String((9, 2), default="O33")
    message_label = HL7String((9, 3), default="OML_O33")
    control_id = HL7String(10)
    processing_id = HL7String(11, default="P")
    version_id = HL7String(12, default="2.5")
    accept_type = HL7String(15, default="NE")
    application_type = HL7String(16, default="AL")
    character_set = HL7String(18, default="ASCII")

    patients = HL7Foreign("PID", "header")
    specimens = HL7Foreign("SPM", "header")
    orders = HL7Foreign("OBR", "header")
    software = HL7Foreign("SFT", "header")
    queries = HL7Foreign("QPD", "header")


class Au800CytologieSegmentSFT(Segment):
    software_version = HL7String(2)
    sofware_product_name = HL7String(3)
    sofware_binary_id = HL7String(4)

    patients = HL7Foreign("PID", "software")


class Au800CytologieSegmentPID(Segment):
    patient_id = HL7String(4)
    patient_name = HL7String(6)

    orders = HL7Foreign("OBR", "patient")
    infos = HL7Foreign("NTE", "patient")
    specimens = HL7Foreign("SPM", "patient")

    def to_string(self, separators=None):
        self.rang_segment = ""
        return super(Au800CytologieSegmentPID, self).to_string(separators=separators)


class Au800CytologieSegmentSPM(Segment):
    specimen_id = HL7String(3)
    specimen_type = HL7String((5, 1))

    # Champ facultatif
    specimen_label = HL7String((5, 2))
    # Champ facultatif
    specimen_loing = HL7String((5, 3))
    specimen_role = HL7String((12, 1))
    specimen_role_label = HL7String((12, 2))
    specimen_role_loing = HL7String((12, 3), necessary=True)

    orders = HL7Foreign("OBR", "patient")
    infos_samples = HL7Foreign("SAC", "patient")
    common_order = HL7Foreign("ORC", "patient")

    def to_string(self, separators=None):
        self.rang_segment = "1"
        return super(Au800CytologieSegmentSPM, self).to_string(separators=separators)


class Au800CytologieSegmentSAC(Segment):
    rack = HL7String(11)
    position = HL7String(12)

    orders = HL7Foreign("OBR", "patient")
    common_order = HL7Foreign("ORC", "patient")
    inventory = HL7Foreign("INV", "patient")

    # Pas de rang de segment pour le segment SAC
    def to_string(self, separators=None):
        self.rang_segment = ""
        return super(Au800CytologieSegmentSAC, self).to_string(separators=separators)


class Au800CytologieSegmentINV(Segment):
    substance_status = HL7String(3)
    expiration_date = HL7Date(13)
    manufacturer_lot_number = HL7String(17)


class Au800CytologieSegmentOBR(Segment):
    specimen_id = HL7String(3)
    universal_service_id = HL7String(5)
    # F: Final results
    # X: if there was an error which prevented a result from being reported
    result_status = HL7String(26)

    common_order = HL7Foreign("ORC", "order")
    infos = HL7Foreign("NTE", "order")
    results = HL7Foreign("OBX", "order")

    def to_string(self, separators=None):
        self.rang_segment = "1"
        return super(Au800CytologieSegmentOBR, self).to_string(separators=separators)


class Au800CytologieSegmentORC(NoRangSegment):
    order_control_code = HL7String(2)
    sample_id = HL7String(4)
    order_status = HL7String(6)
    username_entered = HL7String(11)
    username_validated = HL7String(12)

    results = HL7Foreign("OBX", "order")

    def to_string(self, separators=None):
        self.rang_segment = "NW"
        return super(Au800CytologieSegmentORC, self).to_string(separators=separators)


class Au800CytologieSegmentNTE(Segment):
    comment_text = HL7String(4)
    comment_type = HL7String(5)


class Au800CytologieSegmentTCD(Segment):
    universal_service_identifier = HL7String(2)


class Au800CytologieSegmentQPD(Segment):
    query_tag = HL7String(3)
    user_parameter = HL7String(4)
    specimen_id = HL7String(5)

    info_query = HL7Foreign("RCP", "queries")


class Au800CytologieSegmentRCP(Segment):
    query_priority = HL7String(2)
    quantity_limited_request = HL7String(3)


class Au800CytologieSegmentOBX(Segment):
    value_type = HL7String(3)
    observation_identifier = HL7String((4, 1))
    description = HL7String((4, 2))
    loing_code = HL7String((4, 3))
    value = HL7String(6)
    units = HL7String(7)
    interpretation_code = HL7String(9)
    result_status = HL7String(12)
    observer = HL7String(17)
    analyzer = HL7String(19)
    date = HL7Date(20)

    infos = HL7Foreign("NTE", "result")
    infos_tests = HL7Foreign("TCD", "order")


class Driver(AstmMessageV2):
    SPEC = {
        "code": "atellica800_cytologie",
        "label": "Atellica AU800 HL7 Cytologie Urinaire"
    }
    SEGMENT_CLASS_MAP = {
        "MSH": Au800CytologieSegmentMSH,
        "SFT": Au800CytologieSegmentSFT,
        "PID": Au800CytologieSegmentPID,
        "OBR": Au800CytologieSegmentOBR,
        "SPM": Au800CytologieSegmentSPM,
        "SAC": Au800CytologieSegmentSAC,
        "ORC": Au800CytologieSegmentORC,
        "NTE": Au800CytologieSegmentNTE,
        "TCD": Au800CytologieSegmentTCD,
        "OBX": Au800CytologieSegmentOBX,
        "INV": Au800CytologieSegmentINV,
        "QPD": Au800CytologieSegmentQPD,
        "RCP": Au800CytologieSegmentRCP,
    }
    DEFAULT_SEPARATORS = "|^~\\&"
    FIRST_SEP_INDEX = 3
    SEGMENT_ENDING = "\r"

    OPTIONS = [
        AstmMessageV2.Option(
            "au800_cristaux",
            label="Liste des cristaux séparés par des \";\"",
            default="CRY;.CaOx;.CaOxm;.CaOxd;.TRI;.URI;.CapH;.U-AMO;.P-AMO;.CYS;.LEU;.TYR;.ATY"
        ),
        AstmMessageV2.Option(
            "au800_cylindres",
            label="Liste des cylindres séparés par des \";\"",
            default="HYA;PAT;.C-HGR;.C-GRA;.C-NEC;.C-RBC;.C-WBC;.C-CRY;.C-MIC;.C-FAT;.C-WAX;.C-MIX"
        ),
    ]

    def __init__(
            self,
            au800_cristaux="CRY;.CaOx;.CaOxm;.CaOxd;.TRI;.URI;.CapH;.U-AMO;.P-AMO;.CYS;.LEU;.TYR;.ATY",
            au800_cylindres="HYA;PAT;.C-HGR;.C-GRA;.C-NEC;.C-RBC;.C-WBC;.C-CRY;.C-MIC;.C-FAT;.C-WAX;.C-MIX",
            **kwargs
    ):
        self.au800_cristaux = au800_cristaux
        self.au800_cylindres = au800_cylindres
        super(Driver, self).__init__(**kwargs)

    def create(self, **kwargs):
        return Au800CytologieSegmentMSH(type_segment="MSH", **kwargs)

    def to_string(self, message, separators=None, **kwargs):
        return super(Driver, self).to_string(message, separators, **kwargs) + "\r"

    def encode_worklist(self, worklist, context):
        msg = self.create(
            date=context["now"],
            control_id=context.get(
                "control_id", int(time.time() * 1000)),
        )
        for patient in worklist.get_astm_worklist(
                unique_together=["action", "priority"],
                priority={"A": "A", "S": "S", worklist.ELSE: "R"},
                action={"del": worklist.IGNORE, worklist.ELSE: "G"},
        ):
            patient_name = "{} {}".format(
                patient.last_name,
                patient.first_name,
            )
            seg_pid = msg.patients.add(
                patient_id=patient.patient_id,
                patient_name=patient_name,
            )
            for sample, action, priority in patient.samples:
                for order in sample["orders"]:
                    seg_spm = seg_pid.specimens.add(
                        specimen_id=sample.sample_id,
                        specimen_type="UR",
                        specimen_label="Urine",
                        specimen_loing="HL70487",
                        specimen_role="P",
                        specimen_collection_date=sample.collection_date,
                        specimen_received_date=sample.request_date,
                    )
                    seg_spm.common_order.add(order_control_code="NW")
                    seg_spm.orders.add(specimen_id=sample.sample_id)
        if len(msg.patients):
            return msg
        elif context["query"]:
            for sample_id in [seg_qpd.barcode for seg_qpd in context["query"].queries]:
                seg_pid = msg.patients.add()
                seg_spm = seg_pid.specimens.add(specimen_id=sample_id)
                seg_sac = seg_spm.infos_samples.add(sample_barcode=sample_id)
                seg_sac.common_order.add(order_control_code="DC")
            return msg

    def decode_message(self, message):
        if message.queries:
            worklist = Worklist(context="query")
            for seg_qpd in message.queries:
                sample = worklist.get_or_create(
                    type="patient",
                    sample_id=seg_qpd.specimen_id
                )
        else:
            worklist = Worklist(context="result")
            chemistrery = False
            for seg_sft in message.software:
                if "CLINITEK" in seg_sft.software_version:
                    chemistrery = True
                for seg_pid in seg_sft.patients:
                    for seg_spm in seg_pid.specimens:
                        if seg_spm.specimen_role == "Q":
                            sample = worklist.get_or_create(
                                sample_id=seg_spm.specimen_id,
                                type="control"
                            )
                            for seg_sac in seg_spm.infos_samples:
                                for seg_obr in seg_sac.orders:
                                    for seg_nte in seg_obr.infos:
                                        if seg_nte.comment_type == "RE" and "Lot ID:" in seg_nte.comment_text:
                                            sample.control_lot = seg_nte.comment_text.split(":")[1].strip()
                                    for seg_obx in seg_obr.results:
                                        sample.results.add(
                                            code=seg_obx.observation_identifier if chemistrery else seg_obx.description,
                                            value=seg_obx.value,
                                            date=seg_obx.date,
                                            operator=seg_obx.observer,
                                            instrument=seg_obx.analyzer,
                                            unit=seg_obx.units
                                        )
                        else:
                            sample = worklist.get_or_create(
                                sample_id=seg_spm.specimen_id,
                                specimen=seg_spm.specimen_type,
                                type="patient"
                            )
                            lots = []
                            for seg_sac in seg_spm.infos_samples:
                                for seg_inv in seg_sac.inventory:
                                    lots = [{
                                        "type": "reagent",
                                        "lot": seg_inv.manufacturer_lot_number,
                                        "expiration": seg_inv.expiration_date
                                    }]
                                for seg_obr in seg_sac.orders:
                                    cristaux = []
                                    cylindres = []
                                    for seg_obx in seg_obr.results:
                                        flags = []
                                        if (
                                            seg_obx.description and
                                            seg_obx.description in self.au800_cristaux.split(";") and
                                            seg_obx.value
                                        ):
                                            # Traitement des cristaux
                                            cristaux.append("{}{}".format(seg_obx.description, seg_obx.value))
                                        elif (
                                            seg_obx.description and
                                            seg_obx.description in self.au800_cylindres.split(";") and
                                            seg_obx.value
                                        ):
                                            # Traitement des cylindres
                                            cylindres.append("{}{}".format(seg_obx.description, seg_obx.value))
                                        else:
                                            if seg_obx.interpretation_code != "N":
                                                flags.append(seg_obx.interpretation_code)
                                            # Traitement des autres résultats
                                            if not seg_obx.units:
                                                code = "{}QUAL".format(
                                                    seg_obx.observation_identifier if chemistrery else seg_obx.description,
                                                )
                                            else:
                                                code = seg_obx.observation_identifier if chemistrery else seg_obx.description
                                            sample.results.add(
                                                code=code,
                                                value=seg_obx.value,
                                                flags=flags,
                                                date=seg_obx.date,
                                                operator=seg_obx.observer,
                                                instrument=seg_obx.analyzer,
                                                unit=seg_obx.units,
                                                lots=lots
                                            )
                                    if cristaux:
                                        sample.results.add(
                                            code="CRISTAUX",
                                            value=";".join(cristaux)
                                        )
                                    if cylindres:
                                        sample.results.add(
                                            code="CYLINDRES",
                                            value=";".join(cylindres)
                                        )
        return worklist
