# -*- coding: utf8 -*-
from __future__ import unicode_literals

from datetime import datetime

from driverng import Worklist
from driverng.astm import (
    AstmMessageV2,
    HeaderSegment,
    Segment,
    HprimDate,
    HprimString,
    HprimForeign,
    HprimList,
    HprimMultiple,
)


class TecanEvolutionSegmentH(HeaderSegment):
    sender_name = HprimString((5, 1))
    sender_version = HprimString((5, 2))
    comment = HprimString(11)
    processing_id = HprimString(12, default="P")
    date = HprimDate(14)

    patients = HprimForeign("P", "patients")
    queries = HprimForeign("Q", "queries")
    end = HprimForeign("L", "header", one2one=True, required=True)


class TecanEvolutionSegmentP(Segment):
    # Control Id for QC
    patient_id = HprimString(3)
    field4 = HprimString(4)
    field5 = HprimString(5)

    orders = HprimForeign("O", "orders")


class TecanEvolutionSegmentO(Segment):
    patient_id = HprimString(3)
    specimen_id = HprimString(4)
    assays = HprimList(5, basefield=HprimMultiple(
        0,
        assay=HprimString(4),
        dilution=HprimString(5)
    ))
    action_code = HprimString(12)

    results = HprimForeign("R", "results")


class TecanEvolutionSegmentR(Segment):
    test_id = HprimString((3, 4))
    dilution = HprimString((3, 5))
    result_type = HprimString((3, 8))
    data = HprimString(4)
    units = HprimString(5)
    sample_status = HprimString(11)
    operator_id = HprimString(13)
    date_started = HprimDate(14)
    date_completed = HprimString(15)
    instrument_id = HprimString(16)


class TecanEvolutionSegmentQ(Segment):
    specimen_id = HprimString(3)
    universal_test_id = HprimString((4, 4))


class TecanEvolutionSegmentL(Segment):
    terminaison_code = HprimString(3, default="N")


class Driver(AstmMessageV2):

    SPEC = {
        "code": "tecan_evolution",
        "label": "TECAN EVOLUTION/ANALYSER ASTM",
    }
    SEGMENT_CLASS_MAP = {
        "H": TecanEvolutionSegmentH,
        "P": TecanEvolutionSegmentP,
        "O": TecanEvolutionSegmentO,
        "R": TecanEvolutionSegmentR,
        "Q": TecanEvolutionSegmentQ,
        "L": TecanEvolutionSegmentL,
    }
    DEFAULT_SEPARATORS = "|\\^&"

    def create(self, **kwargs):
        message = TecanEvolutionSegmentH(type_segment="H", **kwargs)
        message.end = TecanEvolutionSegmentL(type_segment="L")
        return message

    def encode_worklist(self, worklist, context):
        message = self.create(
            date=context["now"],
        )
        for patient in worklist.get_astm_worklist(
            unique_together=["action"],
            action={"del": worklist.IGNORE, worklist.ELSE: ""},
        ):
            seg_p = message.patients.add(
                patient_id=patient.patient_id,
                field4=patient.patient_id,
                field5=patient.patient_id,
            )
            for sample, action in patient.samples:
                seg_o = seg_p.orders.add(
                    patient_id=sample.sample_id,
                )
                for order in sample.orders:
                    test = seg_o.assays.add()
                    test.assay = order.code
                    test.dilution = order.dilution
        return message

    def decode_message(self, message):
        if message.queries:
            worklist = Worklist(context="query")
            for seg_q in message.queries:
                worklist.get_or_create(
                    sample_id=seg_q.specimen_id,
                    type="patient",
                )
            return worklist
        worklist = Worklist(context="result")
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                if seg_o.action_code == "Q":
                    sample = worklist.get_or_create(
                        control_lot=seg_p.patient_id,
                        sample_id=seg_p.patient_id,
                        type="control",
                    )
                else:
                    sample = worklist.get_or_create(
                        sample_id=seg_o.specimen_id,
                        type="patient"
                    )
                    for seg_r in seg_o.results:
                        if seg_r.date_completed and len(seg_r.date_completed) > 19:
                            date_result = datetime.strptime(seg_r.date_completed[0:19], "%Y-%m-%dT%H:%M:%S")
                        else:
                            date_result = message.date
                        flags = []
                        # Le status du résultat "Pass" signifie "Success"
                        # Le status "Fail" indique que le résultat est en
                        # erreur
                        if seg_r.sample_status == "Fail":
                            flags.append(seg_r.sample_status)
                        sample.results.add(
                            code="{}{}".format(seg_r.test_id, seg_r.result_type),
                            value=seg_r.data,
                            unit=seg_r.units,
                            dilution=seg_r.dilution,
                            operator=seg_r.operator_id,
                            instrument=seg_r.instrument_id,
                            date=date_result,
                            flags=flags,
                        )
        return worklist
